# -*- coding: utf-8 -*-

"""
    ramlient
    ~~~~~~~~

    Access to a RAML API done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""


class RamlientError(Exception):
    """
        Base Exception for all ramlient based Exceptions
    """
    pass


class ResourceNotFoundError(RamlientError):
    """
        Exception which is raised if a certain resource does not exist
    """
    def __init__(self, resource_path):
        super(ResourceNotFoundError, self).__init__(
            'No such Resource found: {0}'.format(resource_path))


class UnsupportedHTTPMethodError(RamlientError):
    """
        Exception which is raised if a certain HTTP method is not supported
    """
    def __init__(self, method):
        super(UnsupportedHTTPMethodError, self).__init__('No such HTTP method: {0}'.format(method))


class UnsupportedResourceMethodError(RamlientError):
    """
        Exception which is raised if the certain resource does not support the certain method
    """
    def __init__(self, resource_path, method):
        super(UnsupportedResourceMethodError, self).__init__(
            "Resource '{0}' does not support method '{1}'".format(resource_path, method))


class UnsupportedQueryParameter(RamlientError):
    """
        Exception which is raised if the certain resource does not
        support the certain query parameter
    """
    def __init__(self, resource_path, query_parameter):
        super(UnsupportedQueryParameter, self).__init__(
            "Resource '{0}' does not support Query Parameter '{1}'".format(
                resource_path, query_parameter))



# -*- coding: utf-8 -*-

"""
    ramlient
    ~~~~~~~~

    Access to a RAML API done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

import codecs
from collections import namedtuple

import ramlfications

from . import utils
from .request import AVAILABLE_METHODS, prepare_request
from .errors import ResourceNotFoundError, UnsupportedResourceMethodError

NodeParameter = namedtuple("NodeParameter", ["resource", "parameter"])


class Node(object):
    """
        Wrapper for the RAML ResourceNode
    """
    def __init__(self, client, resource):
        self.client = client
        self.resource = resource

    def __repr__(self):
        return "<Node ({0})>".format(repr(self.resource))

    def __getattr__(self, attr):
        """
            Accessing sub node
        """
        if attr in AVAILABLE_METHODS:
            self._patch_resource(attr)
            return prepare_request(self)

        resource = self.client.get_resource(self.resource.path + "/", attr)
        if not resource:
            raise ResourceNotFoundError(self.resource.path + "/" + attr)

        if hasattr(resource, "parameter"):
            return lambda x: ParameterizedNode(
                self.client, resource.resource, {resource.parameter: x})
        else:
            return Node(self.client, resource)

    @property
    def path(self):
        """
            Returns the RAML resource node's path
        """
        return self.resource.path

    def _patch_resource(self, method):
        """
            Patch the current RAML ResourceNode by the resource with the
            correct method if it exists

            If the resource with the specified method does not exist
            an exception is raised.

            :param str method: the method of the resource

            :raises UnsupportedResourceMethodError: if resource does not support the method
        """
        resource = self.client.get_resource("", self.resource.path, method)
        if not resource:
            raise UnsupportedResourceMethodError(self.resource.path, method)

        self.resource = resource


class ParameterizedNode(Node):
    """
        Wrapper for the RAML ResourceNode combined with the values
        of the dynamic part from the URL
    """
    def __init__(self, client, resource, parameter):
        super(ParameterizedNode, self).__init__(client, resource)
        self.parameter = parameter

    def __repr__(self):
        return "<ParameterizedNode ({0}), Parameter: {1}>".format(self.resource, self.parameter)

    @property
    def path(self):
        """
            Returns the RAML resource node's path
        """
        return self.resource.path.format(**self.parameter)


class Client(object):
    """
        RAML client
    """
    def __init__(self, ramlfile, ramlconfig=None):
        self.ramlfile = ramlfile
        self.ramlconfig = ramlconfig
        self.raml = None

        self.parse_raml()
        self.base_uri = self.raml.base_uri

    def parse_raml(self):
        """
            Parse RAML file
        """
        if utils.is_url(self.ramlfile):
            raml = utils.download_file(self.ramlfile)
        else:
            with codecs.open(self.ramlfile, "rb", encoding="utf-8") as raml_f:
                raml = raml_f.read()

        loader = ramlfications.loads(raml)
        config = ramlfications.setup_config(self.ramlconfig)
        self.raml = ramlfications.parse_raml(loader, config)

    def __getattr__(self, attr):
        """
            Access ResourceNode from RAML
        """
        resource = self.get_resource("/", attr)
        if not resource:
            raise ResourceNotFoundError(attr)

        return Node(self, resource)

    def get_resource(self, base_resource_path, resource_path, method=None):
        """
            Gets a resource by it's path and optional by it's method

            This method does not care about the supported resource methods
            unless it is specified.

            :param str resource_path: The path of the resource
            :param str method: The method of the path.

            :returns: the resource if it exists or None
            :rtype: ResourceNode
        """
        basic_path = base_resource_path + resource_path
        dynamic_path = base_resource_path + "{" + resource_path + "}"
        for resource in self.raml.resources:
            method_matched = method is None or resource.method == method
            if method_matched and (resource.path == basic_path
                                   or resource.path == basic_path + '/'):
                return resource

            if resource.path == dynamic_path and method_matched:
                return NodeParameter(resource=resource, parameter=resource_path)
        return None



# -*- coding: utf-8 -*-

"""
    ramlient
    ~~~~~~~~

    Access to a RAML API done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

import os

import pytest

from ramlient import Client


__DATA_DIR__ = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data/'))
__EXAMPLES_DIR__ = os.path.join(__DATA_DIR__, 'examples')
__EXAMPLES_CONFIG_TPL__ = os.path.join(__DATA_DIR__, 'examples', '{name}-config.ini')


@pytest.fixture(scope='function')
def example_client(request):
    """
    Fixture which creates a new client based
    on an example API from tests/data/examples.
    """
    example_name = request.param
    example_path = os.path.join(__EXAMPLES_DIR__, example_name + '.raml')
    example_config = __EXAMPLES_CONFIG_TPL__.format(name=example_name)
    if not os.path.exists(example_config):
        example_config = None

    client = Client(example_path, ramlconfig=example_config)
    return client



# -*- coding: utf-8 -*-

"""
    Test ramlient.utils module.
"""

import pytest

from ramlient import utils


@pytest.mark.parametrize('url, is_url', [
    ('http://foo.com', True),
    ('http://foo.com/bar', True),
    ('http://foo.com/bar#foo', True),
    ('http://foo.com/bar.txt?foo=bar', True),
    ('bar.txt', False),
    ('/usr/bin/', False),
    ('../bar.txt', False),
    ('./bar.txt', False)
])
def test_is_url_function(url, is_url):
    """
        Test if "is_url" function works
    """
    # then
    assert utils.is_url(url) is is_url


@pytest.mark.parametrize('value, expected_type, does_match', [
    ('Hello', 'string', True),
    ('Hello', 'integer', False),
    (42, 'integer', True),
    (42, 'string', False),
    (42, 'weird_type', False),
])
def test_match_type_function(value, expected_type, does_match):
    """
        Test if "match_type" function works
    """
    # then
    assert utils.match_type(value, expected_type) is does_match



# -*- coding: utf-8 -*-

"""
    ramlient
    ~~~~~~~~

    Access to a RAML API done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

import requests
from ramlfications.raml import AVAILABLE_METHODS

from .utils import match_type
from .errors import UnsupportedHTTPMethodError, UnsupportedQueryParameter


def prepare_request(node):
    """
        Prepare request to node's API route

        :param Node node: the RAML node object
    """
    if node.resource.method not in AVAILABLE_METHODS:
        raise UnsupportedHTTPMethodError(node.resource.method)

    def request(data=None, json=None, **kwargs):
        """
            Make request to node's API route with the given keyword arguments
        """
        # validate given query parameters
        for key, value in kwargs.items():
            param = next((p for p in node.resource.query_params if p.name == key), None)
            if not param:
                raise UnsupportedQueryParameter(node.resource.path, key)

            if not match_type(value, param.type):
                raise TypeError(
                    "Resource Query Parameter has type '{0}' but expected type '{1}'".format(
                        value.__class__.__name__, param.type))

        response = requests.request(node.resource.method, node.resource.absolute_uri, params=kwargs,
                                    data=data, json=json)
        return response
    return request



# -*- coding: utf-8 -*-

"""
    ramlient
    ~~~~~~~~

    Access to a RAML API done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

__version__ = '0.1.4'


from .core import Client, Node, ParameterizedNode  # noqa



# -*- coding: utf-8 -*-

"""
    ramlient
    ~~~~~~~~

    Access to a RAML API done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

import sys

try:  # python 2.x
    from urlparse import urlparse
except ImportError:  # python 3.x
    from urllib.parse import urlparse

import requests

if sys.version_info[0] == 2:  # python 2.x
    TYPES_MAPPING = {
        "string": (str, unicode),  # noqa
        "integer": int
    }
else:  # python 3.x
    TYPES_MAPPING = {
        "string": str,
        "integer": int
    }


def download_file(url):
    """
        Downloads a file from the specified URL.

        :param str url: The URL to the file to be downloaded

        :returns: the downloaded file's content
        :rtype: str
    """
    response = requests.get(url)
    if response.status_code is not 200:
        return None
    return response.text


def is_url(url):
    """
        Check if given URL is a valid URL.

        :param str url: The url to validate

        :returns: if the url is valid or not
        :rtype: bool
    """
    return urlparse(url).scheme != ""


def match_type(value, expected_type):
    """
        Checks if the given value type matches the
        expected type

        :param value: the value which's type should be checked
        :param str expected_type: the name of the expected value

        :returns: if the value has the expected type or not
        :rtype: bool
    """
    try:
        return isinstance(value, TYPES_MAPPING[expected_type])
    except KeyError:
        return False



# -*- coding: utf-8 -*-

"""
    Test ramlient.core module
"""

import types

import pytest

from ramlient.core import Node, ParameterizedNode
from ramlient.errors import UnsupportedResourceMethodError, UnsupportedQueryParameter


@pytest.mark.parametrize('example_client', [
    'simple'
], indirect=['example_client'])
def test_creating_client_from_raml_url(example_client):
    """
        Test creating client from a RAML file per URL
    """
    assert example_client.raml.title == 'Example API'


@pytest.mark.parametrize('example_client', [
    'simple'
], indirect=['example_client'])
def test_accessing_first_level_resource(example_client):
    """
        Test accessing first level resource
    """
    # given & when
    resource = example_client.resource

    # then
    assert isinstance(resource, Node)
    assert resource.path == '/resource'
    assert resource.resource.display_name == 'First One'


@pytest.mark.parametrize('example_client', [
    'github'
], indirect=['example_client'])
def test_accessing_second_level_resource(example_client):
    """
        Test accessing second level resource
    """
    # given & when
    resource = example_client.search.repositories

    # then
    assert isinstance(resource, Node)
    assert resource.path == '/search/repositories'


@pytest.mark.parametrize('example_client', [
    'simple'
], indirect=['example_client'])
def test_accessing_dynamic_resource(example_client):
    """
        Test accessing dynamic resource
    """
    # given & when
    resource = example_client.resource.resourceId(5)

    # then
    assert isinstance(resource, ParameterizedNode)
    assert resource.resource.path == '/resource/{resourceId}'
    assert resource.parameter == {"resourceId": 5}
    assert resource.path == '/resource/5'


@pytest.mark.parametrize('example_client', [
    'simple'
], indirect=['example_client'])
def test_getting_request_method_from_resource(example_client):
    """
        Test getting request method from resource
    """
    assert isinstance(example_client.resource.get, types.FunctionType)
    assert isinstance(example_client.resource.put, types.FunctionType)
    assert isinstance(example_client.resource.delete, types.FunctionType)
    assert isinstance(example_client.resource.patch, types.FunctionType)
    assert isinstance(example_client.resource.options, types.FunctionType)
    assert isinstance(example_client.resource.trace, types.FunctionType)
    assert isinstance(example_client.resource.connect, types.FunctionType)

    # when
    with pytest.raises(UnsupportedResourceMethodError) as exc:
        example_client.resource.post  # noqa

    # then
    assert str(exc.value) == "Resource '/resource' does not support method 'post'"

    # then
    assert isinstance(example_client.resource.resourceId(5).get, types.FunctionType)
    assert isinstance(example_client.resource.resourceId(5).post, types.FunctionType)

    # when
    with pytest.raises(UnsupportedResourceMethodError) as exc:
        example_client.resource.resourceId(5).delete  # noqa

    # then
    assert str(exc.value) == "Resource '/resource/{resourceId}' does not support method 'delete'"


@pytest.mark.parametrize('example_client', [
    'simple'
], indirect=['example_client'])
def test_passing_query_parameter_to_resource(example_client):
    """
        Test passing query parameter to resource
    """
    # when
    with pytest.raises(UnsupportedQueryParameter) as exc:
        example_client.resource.resourceId(5).get(foo=42)

    # then
    assert str(exc.value) == "Resource '/resource/{resourceId}' does " \
        "not support Query Parameter 'foo'"


@pytest.mark.parametrize('example_client', [
    'simple'
], indirect=['example_client'])
def test_passing_wrong_typed_query_parameter_to_resource(example_client):
    """
        Test passing wrong typed query parameter to resource
    """
    # when
    with pytest.raises(TypeError) as exc:
        example_client.resource.resourceId(5).get(filter=42)

    # then
    assert str(exc.value) == "Resource Query Parameter has type 'int' but expected type 'string'"
